/* abs2xx.c - Create an 'XX' Learn String from one or more absolute files.
 * by esde 2018
 *
 * The XX Learn String is used by HP 163x series logic analyzers to load
 * Inverse Assemblers and other native programs onto the instrument via
 * HP-IB.
 *
 * /!\ Please note!
 *
 * - It is assumed that files will be specified in the order that they
 *   are expected to appear in the output.  In the average case for an
 *   inverse assembler, this order would be: loader, probe config, and
 *   finally ia.
 *
 * - This program will use the first PIR (Program Information Record)
 *   encountered for the learn string.  Subsequent PIRs will be ignored.
 *
 * - This program does not respect load address high word (always 0)
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <arpa/inet.h>

#include "hpabs.h"

struct xxhdr {
	uint8_t sig[2];
	uint16_t length;
};

void usage() {
	printf("usage: enframe <input1, input2, ...> <output>\n");
	exit(1);
}

int fetch_record(uint8_t *dest, FILE *fp) {
	struct datarec *rec = (struct datarec *) dest;
	int ret, sz, nb;

	nb = ret = fread(rec, sizeof(struct datarec), 1, fp);
	if(ret != sizeof(struct datarec)) {
		printf("Error while reading record header.\n");
		exit(1);
	}

	/* Fix up endianess */
	sz = ntohs(rec->nbytes);

	/* Read rest of record */
	ret = fread(dest+sizeof(struct datarec), 1, sz, fp);
	if(ret != sz) {
		printf("Error while reading record data.\n");
		exit(1);
	}

	return ret + nb;
}

int write_block(uint8_t *buf, int size, FILE *fp) {
	uint8_t wl, checksum = 0;
	int i;

	/* Calculate checksum */
	for(i=0;i<size;i++)
		checksum += buf[i];

	/* Caclulate word length (rounded up to nearest word) */
	wl = (size / 2) + (size % 2);

	fwrite(&wl, sizeof(uint8_t), 1, fp);
	fwrite(buf, 1, size, fp);
	fwrite(&checksum, sizeof(uint8_t), 1, fp);

	return size+2;
}

int main(int argc, char **argv) {
	struct pirec pir;
	struct datarec rec;
	struct xxhdr xx;

	int an, size, wbytes = 0;

	FILE *fpo, *fpi;
	int ret, recs = 0;
	uint8_t buffer[256];

	if(argc < 3)
		usage();

	/* Open output (last argument) */
	fpo = fopen(argv[argc-1], "wb");
	if(fpo == NULL) {
		printf("Error opening output\n");
		exit(1);
	}

	/* Skip XX header- to be filled in later. */
	fseek(fpo, sizeof(xx), SEEK_SET);

	/* Loop through input files */
	for(an=1;an<(argc-1);an++) {

		/* Open input file */
		fpi = fopen(argv[an], "rb");
		if(fpi == NULL) {
			printf("Error opening input\n");
			exit(1);
		}

		/* Read PIR, only write for first file */
		fread(&pir, sizeof(pir), 1, fpi);
		if(an == 1) {
			wbytes += write_block((uint8_t *) &pir, sizeof(pir), fpo);
			recs++;
		}

		while((ret = fread(&rec, sizeof(rec), 1, fpi)) > 0) {
			size = ntohs(rec.nbytes);

			/* XXX: eh. */
			memcpy(buffer, &rec, sizeof(rec));

			ret = fread(&buffer[sizeof(rec)], 1, size, fpi);
			if(ret != size) {
				printf("Encoutered truncated record in input.\n");
				exit(1);
			}

			wbytes += write_block(buffer, size + sizeof(rec), fpo);
			recs++;
		}

		fclose(fpi);
	}

	/* write trailer */
	fwrite("\0\0\0\0", 1, 4, fpo);
	wbytes += 4;

	/* update header */
	xx.sig[0] = 'X';
	xx.sig[1] = 'X';
	xx.length = htons(wbytes);
	rewind(fpo);
	fwrite(&xx, sizeof(xx), 1, fpo);

	fclose(fpo);

	printf("abs2xx: Wrote %d bytes (%d block%s) to %s\n",
		wbytes, recs, recs > 1 ? "s" : "", argv[argc-1]);

	return 0;
}
