/* bin2abs.c - Create an 'HP Absolute' file from a provided binary
 * by esde 2018
 *
 * HP Absolute files are described in chapter 6 of the "HP 64000 Logic
 * Development System - File Format Reference Manual".  This document
 * is available from bitsavers.
 *
 * This, in addition to another framing structure is used by the XX
 * Learn String of HP 163x series Logic Analyzers.
 *
 * /!\ Please note!
 *
 * - This program generates a default PIR which assumes a bus with of 8,
 *   a data width of 8, and a transfer address of 0x3000.
 *
 * - This program does not respect load address high word (always 0)
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <arpa/inet.h>

#include "hpabs.h"

void usage() {
	printf("usage: bin2abs output address file\n");
	exit(1);
}

int write_record(uint8_t *buf, int addr, int size, FILE *fp) {
	struct datarec rec;

	/* build record header */
	rec.nbytes = htons(size);
	rec.addrl = htons(addr);
	rec.addrh = 0;

	/* write data */
	fwrite(&rec, sizeof(rec), 1, fp);
	fwrite(buf, size, 1, fp);

	/* if needed, emit an extra pad byte */
	if(size % 2)
		putc('\0', fp);

	return size;
}

int main(int argc, char **argv) {
	struct pirec pir;
	FILE *fpo, *fpi;
	int ret, recs = 0;
	uint16_t address;
	uint8_t buffer[256];

	if(argc < 4)
		usage();

	/* Get address argument. */
	address = strtol(argv[2], NULL, 16);

	/* Open input, allowing for stdin. */
	if(!strcmp(argv[3], "-")) {
		fpi = stdin;
	} else {
		fpi = fopen(argv[3], "rb");
		if(fpi == NULL) {
			printf("Error opening input\n");
			exit(1);
		}
	}

	/* Open output */
	fpo = fopen(argv[1], "wb");
	if(fpo == NULL) {
		printf("Error opening output\n");
		exit(1);
	}

	/* Write Processor Information Record */
	pir.buswidth = htons(0x0008);
	pir.datawidth = htons(0x0008);
	pir.taddrl = htons(address);
	pir.taddrh = 0;
	fwrite(&pir, 1, sizeof(pir), fpo);

	/* Read in blocks of a maximum of 250 bytes and keep doing so
     * while input remains available.
	*/
	while((ret = fread(buffer, 1, 0xfa, fpi)) > 0) {
		write_record(buffer, address, ret, fpo);
		address += ret;
		recs++;
	}

	fclose(fpo);

	/* Close input if it was from a file. */
	if(fpi != stdin)
		fclose(fpi);

	printf("bin2abs: Wrote %d record%s to %s\n", recs, recs > 1 ? "s" : "", argv[1]);

	return 0;
}
