/* reframe.c - Check and optionally update checksums in an 'XX' Learn String
 * by esde 2018
 *
 * This attempts to scan the blocks in a given file, and update the block
 * checksums.  This allows for a more comfortable direct manipulation of
 * an XX string (in a hex editor, for example).  Some attempt is made to
 * check that the input file is sane.
 *
 * /!\ Please note!
 *
 * - Although some precautions have been taken, running this tool on
 *   malformed input could be destructive.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <unistd.h>

/* Conditional string macro...maybe a slight readbility improvment */
#define CS(x, s) (x) ? s : ""

struct xxhdr {
	uint8_t sig[2];
	uint16_t length;
};

void usage() {
	printf("usage: reframe [-n] <file>\n");
	exit(1);
}

void bail(char *fmt, ...) {
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	exit(1);
}

int main(int argc, char **argv) {
	FILE *fp;
	struct xxhdr xx;
	int an = 1, i, bsize, nbytes, ret, update_flag = 1;
	uint8_t wsize, checksum, calcsum;
	uint8_t buffer[256];

	int endflag = 0, nblks = 0, nupdates = 0;
	struct stat st;

	if(argc < 2)
		usage();

	if(!strcmp(argv[an], "-n")) {
		update_flag = 0;
		an++;
	}

	/* Get input stats */
	if(stat(argv[an], &st) == -1)
		bail("Error doing stat.\n");

	/* Open output */
	fp = fopen(argv[an], "rb+");
	if(fp == NULL)
		bail("Error opening output.\n");

	/* Read header */
	if(fread(&xx, sizeof(xx), 1, fp) < 1)
		bail("Error reading file header.\n");

	if(xx.sig[0] != 'X' || xx.sig[1] != 'X')
		bail("File does not appear to be an 'XX' Learn String.\n");

	/* Get file length */
	nbytes = ntohs(xx.length);

	/* Check file sizes */
	/* Maybe some more elaborate checks? Specify bigger/smaller? Absurd sizes? */
	if(nbytes + 4 != st.st_size)
		printf("WARNING: File size and header size do not agree (f 0x%lx, h 0x%x).\n", st.st_size, nbytes);

	/*
	 * Read and check blocks
	*/
	while((ret = fread(&wsize, 1, 1, fp)) == 1) {
		if(wsize > 0x80)
			bail("Invalid block length encountered %x.\n", wsize);

		if(wsize == 0) {
			endflag = 1;
			break;
		}

		/* Convert to bytes */
		bsize = wsize * 2;

		/* Check if there are sufficient bytes remaining */
		if(bsize + 1 > nbytes)
			bail("Block length extends beyond end of file.\n");

		if(fread(buffer, 1, bsize, fp) != bsize)
			bail("Error while reading block data.\n");

		if(fread(&checksum, 1, 1, fp) != 1)
			bail("Error while reading checksum.\n");

		nblks++;

		/* Calculate current checksum */
		calcsum = 0;
		for(i=0;i<bsize;i++)
			calcsum += buffer[i];

		/* Compare and update sum */
		if(calcsum != checksum) {
			if(update_flag) {
				fseek(fp, -1, SEEK_CUR);
				fwrite(&calcsum, 1, 1, fp);
			}
			nupdates++;
		}

		/* Subtract consumed bytes from total (block size + checksum + word length) */
		nbytes -= (bsize + 2);
	}

	fclose(fp);

	if(nbytes > 0x20)
		printf("WARNING: Excessive bytes (%d) remaining.\n", nbytes);

	if(!endflag)
		printf("WARNING: File possibly ended prematurely.\n");

	if(nupdates) {
		printf("reframe: %sUpdated %d checksum%s out of %d block%s in %s\n",
			CS(!update_flag, "(dry-run) "), nupdates, CS(nupdates > 1, "s"),
			nblks, CS(nblks > 1, "s"), argv[an]);
	} else
		printf("reframe: %d block%s checked OK in %s\n",
			nblks, CS(nblks > 1, "s"), argv[an]);

	return 0;
}
