/* reloc.c - A tool to produce an absolute file from an HP Relocatable.
 * by esde 2018
 *
 * 13nov13 - (prehistory)
 *  6feb18 - original burst
 * 26sep18 - added argument handling,
 *           added symbol relocation format 3 (absolute offset to segment)
 * 28sep18 - partial/good-enough fix for name record handling
 *           implemented merged output array (each seg used to have its own)
 *
 * TODO:
 *  [ ] ensure NAME record occurs before further processing
 *  [ ] possibly break symbol relocations into their own function
 *  [ ] expanded argument handling to support extra modes (like debug)
 *  [ ] maybe work up a log facility so quiet runs would be possible
 *  [ ] rework the 'dest' pointer thing in parseTBlock.  it's not really
 *      needed anymore i dont think now that the absbuf is passed in.
 *  [ ] consider a set of functions for writing to output array.  this
 *      would then allow for more robust checking of overwrites and so on.
 *
 * The format of the HP Relocatable is described in the "File Format Reference
 * Manual" for the HP 64000 Logic Development System.  Manual p/n 64980-90933
 * available from bitsavers.

 * It should be noted that the HP64000 and HP163x segment names differ. This
 * code is generally using the former, but it is a little confusing.  To
 * clarify: this chart...
 *
 *   0  ABS    Absolute/Setup/Config
 *   1  PROG   Variables
 *   2  DATA   Program code
 *   3  COMN   Constats
*/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <endian.h>
#include <ctype.h>
#include <string.h>

/*
 * Helpful color macros
*/
#define grn(x) "\033[32m"x"\033[0m"
#define red(x) "\033[31m"x"\033[0m"
#define yel(x) "\033[33m"x"\033[0m"

/* Used by certain forms of relocation
 * XXX: Ideally wouldn't need to be global...?
*/
uint16_t loadAddress;

#define SEG_ABS	  0
#define SEG_PROG  1
#define SEG_DATA  2
#define SEG_COMN  3

struct absbuf {
	uint8_t *buf;			/* Space allocated for output buffer */
	uint8_t *seg[4];		/* Array of pointers to segment starts in buf */
	uint16_t len[4];		/* Length of segments (for convenience) */
};

/* "classic" memory dump */
void u_prettydump(unsigned char *buf, int offs, int l) {
    int o = offs, r = l, a, i;
    char tmp[17];

    while(r > 0) {
        a = (r / 16) ? 16 : r % 16;

        printf("0x%.5x  ", o);

        for(i=0; i<a; i++) {
            int q = o+i;
            printf("%.2x ", buf[q]);
            tmp[i] = isprint(buf[q]) ? buf[q] : '.';
        }
        tmp[a] = '\0';

        printf("%*s %s\n", 48 - a*3, "", tmp);

        o += 16;
        r -= a;
    }
}

/*
 * Checksum routines
*/
uint16_t checksum = 0;

void csStart(uint16_t in) {
	checksum = in;
}

void csReset() {
	checksum = 0;
}

uint16_t csFetch(uint16_t *buf) {
	uint16_t val = be16toh(*buf);

	checksum += val;
	return val;
}

int csCheck(uint16_t *buf) {
	uint16_t val = be16toh(*buf);

	return val == checksum;
}

/*
 * Block parsers
*/
const char *getTypeStr(int val) {
	if(val == 0)
		return "Absolute (no reloc)";
	else if(val == 1)
		return "Program";
	else if(val == 2)
		return "Data";
	else if(val == 3)
		return "Common";
	else
		return "Unknown";
}

size_t parseName(uint16_t *buf, uint16_t len, struct absbuf *ab) {
	/* BIG XXX: This routine has sort of a lot of issues...
	 *
	 *  - Buffer size allocation is fixed!
	 *
	 *  - Potential memory leak!
	 *
	 *  - It's sort of annoying that the length fields are in words.
	 *
	 *  - Length byte is packed in with the first char of the string.
	 *    This seems in conflict with the 64000 docs, which describes a much
	 *    more elaborate situation than what is observed.
	 *
	 *  - Technically the segment lengths are doubleword in length but that's
	 *    not terribly likely with the inverse assembler relocatables, so
	 *    whatever...but it's bad form.
	 *
	 *  - This is a nasty hack but we'll need these values for symbol
	 *    relocations later, and I'd rather prefer to avoid a second pass.
	 *    At least one of the nasty things about this is that it requires
	 *    that a name record come before any relocation.  I think this will
	 *    be true of output from the IAL assembler, but its sloppy.
	*/

	uint16_t namelen = be16toh(buf[1]) >> 8;

	ab->len[SEG_ABS] = 0;
	ab->len[SEG_PROG] = be16toh(buf[namelen+2]);
    ab->len[SEG_DATA] = be16toh(buf[namelen+4]);
	ab->len[SEG_COMN] = be16toh(buf[namelen+6]);

	/* Allocate space in buffer */
	ab->buf = malloc(16 * 1024);
	memset(ab->buf, 0,  16 * 1024);

	/* Calculate offsets into buffer data...
	 * These will be used later, largely by parseTBlock to do the actual data
	 * placement.  When the block instructs a certain destination segment, its
	 * pointer will be retreived from this array.  Special cases, like the
	 * absolute segment, are based at the start of the data buffer, thereby
	 * allowing them to be placed anywhere within the out data buffer.
	*/
	ab->seg[SEG_ABS] = ab->buf;
	ab->seg[SEG_PROG] = ab->buf;
	ab->seg[SEG_DATA] = ab->buf + ab->len[SEG_PROG]*2;
	ab->seg[SEG_COMN] = ab->buf + ab->len[SEG_PROG]*2 + ab->len[SEG_DATA]*2;

	printf(" Segment lengths (in decimal words)...\n");
	printf("   PROG %d\n", ab->len[SEG_PROG]);
	printf("   DATA %d\n", ab->len[SEG_DATA]);
	printf("   COMN %d\n", ab->len[SEG_COMN]);

	printf("segs: PROG %p, DATA %p, COMN %p\n",
		ab->seg[SEG_PROG], ab->seg[SEG_DATA], ab->seg[SEG_COMN]);

	return 0;
}

/*
 * Parse and handle a full T parameter block.
 * Takes a pointer to the start of the block, a length of remaining
 * words in the entire DOUBLE record, and a destination pointer.
 * Returns the number of words consumed, up to a maximum of 9.
*/
int parseTBlock(uint16_t *buf, int len, uint8_t **dest, struct absbuf *ab) {
	uint16_t *ptr = buf;
	uint8_t *dst = *dest;
	uint16_t tparam, data;
	uint8_t t[8];
	int i;

	tparam = csFetch(ptr++);
	printf("T Block %.4x\n", tparam);

	/* Unpack the T parameters for convenience */
	for(i=7; i>=0; i--) {
		t[i] = tparam & 0x03;
		tparam >>= 2;
	}

	for(i=0; i<8 && len > 1; i++) {
		printf("%.2d - len%.2d -", i, len);
		switch(t[i]) {
			case 0:			/* Single byte absolute */
				data = csFetch(ptr++);
				len--;

				*(dst++) = data & 0xff;

				printf("\033[1;32mSingle\033[0m\n  out %.2x\n", data & 0xff);

				break;
			case 1:			/* Word absolute */

				data = csFetch(ptr++);
				len--;

				*(dst++) = (data >> 8) & 0xff;
				*(dst++) = data & 0xff;

				printf("  out %.4x\n", data);

				break;
			case 2: {		/* Symbol relocation */
				uint16_t p1, p2;
				uint8_t cntr, fmt;
				p1 = csFetch(ptr++);		/* format   */
				p2 = csFetch(ptr++);		/* address  */
				len -= 2;

				cntr = (p1 & 0x0180) >> 7;  /* bits 8,7 */
				fmt = p1 & 0x007f;          /* bits 6,0 */

				/* Ok, this is probably an offset to a segment */
				/* segment order is program,vars(data),common */
				/* (I've always wanted an excuse to use a switch this way) */
				if(fmt == 3) {
					printf(grn(" Symbol relocation") ", format 3...\n");

					// untested alternative...ab->seg[cntr] - ab->buf + p2 * 2 + loadAddress
					uint32_t offset = loadAddress + p2*2;
					switch(cntr) {
						case SEG_COMN:
							offset += ab->len[SEG_DATA] * 2;
						case SEG_DATA:
							offset += ab->len[SEG_PROG] * 2;
						case SEG_PROG:
							break;
						default:
							printf(red("Error") ": Illegal segment specified %d\n", cntr);
					}
					printf("           : Abosolute offset to %s is 0x%.4x\n", getTypeStr(cntr), offset);

					/* emit the thing */
	                *(dst++) = (offset >> 8) & 0xff;
	                *(dst++) = offset & 0xff;
				} else {
					printf("\n" yel("Warning") ": Not performing symbol relocation\n");
					printf("       : p1 %.4x, p2 %.4x\n", p1, p2);
					printf("       : segment %.4x (%s), format %.4x\n", cntr, getTypeStr(cntr), fmt);
				}

				} break;
			case 3:			/* Skeleton */
				printf(yel("Warning") ": Not performing skeleton\n");
				break;
			default:
				printf(red("Error") ": Invalid T Parameter value.  (%d)\n", t[i]);
				return -1;
		}
	}

	*dest = dst;

	/* Return the number of words consumed */
	return ptr - buf;
}

void parseDouble(uint16_t *buf, int len, struct absbuf *ab) {
	uint16_t dstaddr, reloc, addr;
	uint16_t *dptr = buf;
	uint8_t *dest;
	int ret;

	/* Reset checksum accumulator */
	csStart(0);

	/* Get record type, destination address and relocation type */
	csFetch(dptr++);
	addr = csFetch(dptr++);
	reloc = csFetch(dptr++);
	len -= 3;

	printf("Address %.4x (%d)\n", addr, addr);
	printf("Reloc   %.4x %s\n", reloc, getTypeStr(reloc));

	/* Basically, we should read the first three words and then kick
	 * into parsing the tblocks.
	 *
	 * The destination address is specified in words, so we must
	 * calculate the offset in bytes.
	*/
	dstaddr = addr * sizeof(uint16_t);

	dest = &ab->seg[reloc][dstaddr];

	while(len > 0) {
		printf(">dest before %p\n", dest);
		ret = parseTBlock(dptr, len, &dest, ab);
		printf(">dest after %p\n", dest);

		if(ret < 0) {
			printf(red("Error") ": TBlock parse failed\n");
			break;
		}
		printf("TBlock consumed %d words\n", ret);

		len -= ret;
		dptr += ret;
		printf("Data remaining %d\n", len);
	}

	printf("%s\n", csCheck(dptr) ? "Checksum " grn("OK") : red("Bad checksum"));
	printf("---end DOUBLE\n\n");
}

/*
 * Output functions
*/
void writeOutput(char *fn, struct absbuf *ab) {
	int fdo;

	fdo = open(fn, O_CREAT | O_RDWR | O_TRUNC, 0644);
	if(fdo < 0) {
		printf("Error opening output file.\n");
		exit(1);
	}

	write(fdo, ab->seg[SEG_PROG], ab->len[SEG_PROG] * 2);
	write(fdo, ab->seg[SEG_DATA], ab->len[SEG_DATA] * 2);
	write(fdo, ab->seg[SEG_COMN], ab->len[SEG_COMN] * 2);

	close(fdo);
}

void writeOutputDebug(struct absbuf *ab) {
	int fdo;

	fdo = open("common.bin", O_CREAT | O_RDWR | O_TRUNC, 0644);
	write(fdo, ab->seg[SEG_COMN], ab->len[SEG_COMN]*2);
	close(fdo);

	fdo = open("program.bin", O_CREAT | O_RDWR | O_TRUNC, 0644);
	write(fdo, ab->seg[SEG_DATA], ab->len[SEG_DATA]*2);
	close(fdo);

	fdo = open("vars.bin", O_CREAT | O_RDWR | O_TRUNC, 0644);
	write(fdo, ab->seg[SEG_PROG], ab->len[SEG_PROG]*2);
	close(fdo);
}

/*
 * Main
*/
int main(int argc, char **argv) {
	uint16_t record[128];		/* maximum record length is 128 words */
	struct absbuf ab;
	int fd, ret;
	int stop = 0;

	uint16_t length, recType;

	/* Check arguments */
	if(argc < 4) {
		printf("usage:\n  reloc <object.R> <load address> <output.bin>\n");
		exit(1);
	}

	/* Get address argument */
	loadAddress = strtol(argv[2], NULL, 16);

	/* Open input */
	fd = open(argv[1], O_RDONLY);
	if(fd < 0) {
		printf("Error opening %s for read\n", argv[1]);
		exit(1);
	}

	printf("Processing object file '%s' with load address of 0x%.4x\n\n",
		argv[1], loadAddress);

	/* First things first, read the file header and check the magic number */
	ret = read(fd, record, 18);
	if(ret != 18) {
		printf("Error while reading file header\n");
		return 1;
	}

	if(be16toh(record[0]) != 0x8203) {
		printf("Correct magic number not found in header\n");
		return 1;
	}

	/*
	 * Now we should be ready to start reading records:
	 *  Read a buffer
	 *  Parse
	 *  Adjust filepos to bytes consumed by parse
	 *
     * The above method works but is not needed.  The records have a length
	 * element...
	*/

	/* Having consumed the header, iterate through the file's records */
	while(!stop) {
		/* Read in block length...
		 * This is a little odd since 0 indicates a length of 1, and
		 * unlike most other things, this is specified in bytes instead
		 * of words.
		*/
		read(fd, &length, sizeof(uint16_t));
		length = be16toh(length) + 1;
		if(length > 132) {
			printf("Record length is unreasonably large (%d).\n", length);
			break;
		}

		ret = read(fd, record, length);
		if(ret < length) {
			printf("Unable to read expected block length of %d bytes.\n", length);
			break;
		}

		/* Adjust length to word size */
		if(length % sizeof(uint16_t) != 0)
			printf("Warning: uneven length\n");

		length = length / sizeof(uint16_t);

		/* Don't include checksum */
		length -= 1;

		recType = be16toh(*record);
		switch(recType) {
			case 1:
				printf("Found NAME record of length %x words\n", length);
				parseName(record, length, &ab);
				/* so.  here we could construct the output buffer according
				 * to the NAME parameters.  then, a flag would need to be set
				 * so further processing didn't occur without a NAME.
				 *
				 * probably some lookup-array with the segment#->pointer/index
				 * passed in to parseDouble.  pD would need to check if that
				 * table is null...if so, then assume no NAME had been parsed
				 * and bail out...
				*/
				break;
			case 3:
				printf("Found DOUBLE record of length 0x%.2x words\n", length);
				parseDouble(record, length, &ab);
				break;
			case 4:
				printf("Found EXTERNAL record, skipping. (%d length)\n", length);
				break;
			case 5:
				printf("Found END record.\n");
				stop = 1;
				break;
			default:
				printf("Invalid record type encountered %.4x, halting.\n", recType);
				stop = 1;
				break;
		}
	}

	close(fd);

	/* XXX: One more hacky-not-nice thing.  So, the absolute segment kinda
	 *      overlays the rest of the relocated data, most often the
	 *      PROG/Vars segment.  Since I'm handling absolute as a separate
	 *      area, it will need to be manually copied over PROG/Vars.  As it
	 *      is now, only absolutes affecting PROG/Vars will be dealt with.
	 *		In the future, the right way to do all this would probably be
	 *		to combine PROG/DATA/COMN into a single memory region from the
	 *		start (and we probably can since NAME record comes early) and then
	 *		it won't be as hard to populate absolute relocs as they are
	 *      encountered.  Of course this assumes that the segments are always
	 *      combined in this order, but it would seem to be so for IAL stuff
     *      and this tool is probably purpose built enough to accept that
	 *      as a fixed behavior.
	*/

	/* dump out segments */
	printf("---PROG---(Variables)---\n");
	u_prettydump(ab.seg[SEG_PROG], 0, ab.len[SEG_PROG] * 2);

	printf("---DATA---(Program code)---\n");
	u_prettydump(ab.seg[SEG_DATA], 0, ab.len[SEG_DATA] * 2);

	printf("---COMMON---(Constants)---\n");
	u_prettydump(ab.seg[SEG_COMN], 0, ab.len[SEG_COMN] * 2);

	/* write merged output */
	writeOutput(argv[3], &ab);

	if(ab.buf)
		free(ab.buf);

	return 0;
}
